/*
 * Copyright (c) Linux community.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to th e
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
*/
#include <libgnome-media-profiles/gnome-media-profiles.h>
#include "media-profiles.h"
#include "log.h"

// Get media-profiles from the GNOME's media library (via the libgnome-media-dev package).
// You can find these values in the GConf registry.
// Start gconf-editor and browse to: system -> gstreamer -> 0.10 -> audio -> profiles.

// Support Gstreamer 1.0 on all platforms.
// Here are some default audio formats with pipeline fragments.
// We need these until gnome-media adds new profiles for Gstreamer 1.0.
typedef struct {
    gchar *id;
    gchar *name;
    gchar *ext;
    gchar *pipe;
} ProfileRec;

ProfileRec DefaultProfiles[] = {
 {"aac",       "CD Quality, AAC",     "m4a", "audio/x-raw,rate=44100,channels=2 ! faac profile=2 ! ffmux_mp4"},
 {"cdlossless","CD Quality, Lossless","flac","audio/x-raw,rate=44100,channels=2 ! flacenc name=enc"},
 {"cdlossy",   "CD Quality, Lossy",   "ogg", "audio/x-raw,rate=44100,channels=2 ! vorbisenc name=enc quality=0.5 ! oggmux"},
 {"mp2",       "CD Quality, MP2",     "mp2", "audio/x-raw,rate=44100,channels=2 ! twolame name=enc mode=0 bitrate=192 ! id3v2mux"},
 {"mp3",       "CD Quality, MP3",     "mp3", "audio/x-raw,rate=44100,channels=2 ! lamemp3enc name=enc target=0 quality=6 ! xingmux ! id3v2mux"},
 {"voicelossless","Voice, Lossless",  "wav", "audio/x-raw,rate=22050,channels=1 ! wavenc name=enc"},
 {"voicelossy",   "Voice, Lossy",     "spx", "audio/x-raw,rate=32000,channels=1 ! speexenc name=enc ! oggmux"}
};

static gchar *fix_profile_str(const gchar *prof_str);
static gboolean media_profile_test_pipeline(const gchar *pipe);
static GList *gm_audio_profile_get_active_ONLY();

static const ProfileRec *media_profile_get_EXT(gchar *id);
static void media_profiles_add_data_EXT(GtkListStore *store);
static gboolean media_profiles_check_profile_EXT(gchar *id);
static gchar *media_profiles_get_pipeline_EXT(gchar *id);
static gchar *media_profiles_get_extension_EXT(gchar *id);

// Keep a list of supported media profiles (audio formats).
// Notice: If this list is empty then we take default values from DefaultProfiles[] array.
static GList *media_profiles_list = NULL;

void media_profiles_init() {
    LOG_DEBUG("Init media-profiles.c.\n");

    // Init GNOME's media profiles module
    gnome_media_profiles_init(NULL);
}

void media_profiles_exit() {
    LOG_DEBUG("Clean up media-profiles.c.\n");

    // Clean up. Free the list, do not touch data items.
    if (media_profiles_list) {
        g_list_free(media_profiles_list);
    }
    media_profiles_list = NULL;
}

static void media_profiles_load() {
    // Load _active_ media profiles (audio formats)
    if (media_profiles_list) return;
    media_profiles_list = gm_audio_profile_get_active_ONLY();
}

static gboolean media_profile_test_pipeline(const gchar *pipe) {
    // Test if system has support for given audio format. 
    // Create and test a basic pipeline.

   // Fix for Gstreamer 1.0
    gchar *fixed_pipe = fix_profile_str(pipe);

    // Pipeline fragment
    gchar *test = g_strdup_printf ("fakesrc ! %s ! fakesink", fixed_pipe);

    gboolean pipe_OK = FALSE;

    // Parse pipeline
    GError *error = NULL;
    GstElement *p = gst_parse_launch(test, &error);
    if (error) {
        g_error_free(error);
    } else {
        pipe_OK = TRUE;
    }

    if (GST_IS_OBJECT(p)) {
        gst_object_unref(p);
    }

    g_free (test);
    g_free(fixed_pipe);

    return pipe_OK;
} 

static gchar *fix_profile_str(const gchar *prof_str) {
   // In Gstreamer 1.0: 
   // Replace "audio/x-raw-int" and "audio/x-raw-float" with "audio/x-raw".
   // Ref: https://wiki.ubuntu.com/Novacut/GStreamer1.0

   gchar *pos = g_strstr_len(prof_str, 15, "audio/x-raw-");
   gchar *s = NULL;

   if (pos) {
     gchar *comma = g_strstr_len(pos, -1, ",");
     if (comma) {
       s = g_strdup_printf("audio/x-raw%s", comma);
     }
   }

   if (!s) {
     s = g_strdup(prof_str);
   }

   // Caller should g_free() this value
   return s;
}


GList *gm_audio_profile_get_list(void);

static GList *gm_audio_profile_get_active_ONLY() {
  // Return a GList of active GMAudioProfile's only.
  // Copyright: GNOME's libgnome-media-profiles-dev package.
  GList *list = NULL, *orig = NULL;
  GList *new_list = NULL;

  orig = list = gm_audio_profile_get_list();

  new_list = NULL;
  while (list)
  {
    GMAudioProfile *profile;

    profile = (GMAudioProfile *)list->data;

    if (gm_audio_profile_get_active (profile)) {
        const gchar *pipe = gm_audio_profile_get_pipeline(profile);

        // Test pipeline
        gboolean pipe_OK = media_profile_test_pipeline(pipe);

        if (pipe_OK) {
            // It's OK. Add to the list.
            new_list = g_list_prepend(new_list, list->data);
        }
    }

    list = g_list_next(list);
  }

  g_list_free(orig);
  return g_list_reverse(new_list);
}

GMAudioProfile *media_profiles_get_profile(gchar *id) {
    // Return media profile for the given id.
    // See: gconf-editor, key: system -> gstreamer -> 0.10 -> audio -> profiles.
    GMAudioProfile *profile = gm_audio_profile_lookup(id);

    // Do not free this value.
    return profile;
}

gchar *media_profiles_get_extension(gchar *id) {
    // Return file extension (such as .mp3, .ogg, .m4a) for the given profile id.
    // See: gconf-editor, key: system -> gstreamer -> 0.10 -> audio -> profiles (value for extension).
    const gchar *extension = NULL;

    if (g_list_length(media_profiles_list) < 1) {
        // From DefaulProfiles[]
        return media_profiles_get_extension_EXT(id);
    }

    GMAudioProfile *profile = gm_audio_profile_lookup(id);
    if (profile) {
        extension = gm_audio_profile_get_extension(profile);
    }

    // Caller should g_free() this value
    return g_strdup(extension);
}

gchar *media_profiles_get_pipeline(gchar *id) {
    // Return partial GStreamer pipeline for the given id.
    // See: gconf-editor, key: system -> gstreamer -> 0.10 -> audio -> profiles (value for pipeline).

    if (g_list_length(media_profiles_list) < 1) {
        // From DefaulProfiles[]
        return media_profiles_get_pipeline_EXT(id);
    }

    const gchar *pipe_str = NULL;
    GMAudioProfile *profile = gm_audio_profile_lookup(id);
    if (profile) {
        pipe_str = gm_audio_profile_get_pipeline(profile);
    }

    // Fix for Gstreamer 1.0
    gchar *fixed_str = fix_profile_str(pipe_str);

    // Caller should g_free() this value
    return fixed_str;
}

void media_profiles_add_data(GtkListStore *store) {
    // Fill the list store with _active_ media profiles (audio format ids and names).

    media_profiles_load();

    // Did we get media profiles from GConf or DConf-registry (for Gstreamer 1.0)?
    if (g_list_length(media_profiles_list) < 1) {

        // Inform user/developer that we're taking values from DefaulProfiles[] array, 
        // and not from GNOME's GConf or DConf registry. 

        g_print("Notice: Audio-recorder found no audio formats in GConf-registry (for Gstreamer 1.0). "
                "Instead, using hard-coded default values for audio pipelines.\n");

        // From DefaulProfiles[]
        media_profiles_add_data_EXT(store);
    }

    GtkTreeIter iter;

    GList *item = g_list_first(media_profiles_list);
    while (item) {
        GMAudioProfile *profile = (GMAudioProfile*)item->data;

        const gchar *id = gm_audio_profile_get_id(profile);
        const gchar *name = gm_audio_profile_get_name(profile);
        const gchar *extension = gm_audio_profile_get_extension(profile);

        // Typical audio format list is
        //  CD Quality, AAC (.m4a)
        //  CD Quality, Lossless (.flac)
        //  CD Quality, Lossy (.ogg)
        //  CD Quality, MP2 (.mp2)
        //  CD Quality, MP3 (.mp3)
        //  Voice, Lossless (.wav)
        //  Voice, Lossy (.spx)

        gchar *e = g_utf8_strup(extension, -1);
        gchar *n = g_utf8_strdown(name, -1);
        gchar *txt = g_strdup_printf(".%s  (%s)", e, n);

        // Add OGG type "CD Quality, Lossy" as first (becomes default)
        if (!g_strcmp0(extension, "ogg") || !g_strcmp0(extension, "oga")) {
            gtk_list_store_prepend(GTK_LIST_STORE(store), &iter);
        } else {
            // Add last
            gtk_list_store_append(GTK_LIST_STORE(store), &iter);
        }

        gtk_list_store_set(GTK_LIST_STORE(store), &iter, 0, id, 1, txt, -1);

        g_free(txt);
        g_free(e);
        g_free(n);

        item = g_list_next(item);
    }
}


gchar *media_profiles_get_selected_id(GtkWidget *widget) {
    // Return selected GMAudioProfile id (audio format id)

    GtkComboBox *combo = GTK_COMBO_BOX(widget);
    g_return_val_if_fail (GTK_IS_COMBO_BOX(combo), NULL);

    GtkTreeIter iter;
    gchar *id = NULL;

    if (!gtk_combo_box_get_active_iter(combo, &iter)) return NULL;

    gtk_tree_model_get(gtk_combo_box_get_model(combo), &iter, COL_PROFILE_ID, &id, -1);

    // The caller should g_free() this value
    return id;
}

GtkWidget *media_profiles_create_combobox() {
    // Create a GtkComboBox with N_PROFILE_COLUMNS. Populate it with supported audio formats.

    // Create list store
    GtkListStore *store = gtk_list_store_new(N_PROFILE_COLUMNS, G_TYPE_STRING, G_TYPE_STRING);

    // Get audio formats (media profiles) and add to the store
    media_profiles_add_data(store);

    GtkWidget *combo = gtk_combo_box_new_with_model(GTK_TREE_MODEL(store));

    // Unref store
    g_object_unref(store);

    // "Id" column, invisible
    GtkCellRenderer *cell = gtk_cell_renderer_text_new();
    g_object_set(cell, "visible", 0, NULL);
    gtk_cell_layout_pack_start(GTK_CELL_LAYOUT(combo), cell, TRUE);
    gtk_cell_layout_set_attributes(GTK_CELL_LAYOUT(combo), cell, "text", COL_PROFILE_ID, NULL);

    // "Name" column, visible
    cell = gtk_cell_renderer_text_new();
    gtk_cell_layout_pack_start(GTK_CELL_LAYOUT(combo), cell, TRUE);
    gtk_cell_layout_set_attributes(GTK_CELL_LAYOUT(combo), cell, "text", COL_PROFILE_TXT, NULL);

    return combo;
}

gboolean media_profiles_check_profile(gchar *id) {
    // Check if id is a valid profile ident

    if (g_list_length(media_profiles_list) < 1) {
        // DefaulProfiles[]
        return media_profiles_check_profile_EXT(id);
    }

    GMAudioProfile *profile = media_profiles_get_profile(id);
    return (profile != NULL);
}

// ------------------------------------------------------
// Functions to get values from DefaultProfiles[] array
// ------------------------------------------------------

static void media_profiles_add_data_EXT(GtkListStore *store) {
    // Fill the list store with _active_ media profiles (audio format ids and names).
    GtkTreeIter iter;

    // DefaulProfiles[]
    guint i = 0;
    for (i = 0; i < sizeof(DefaultProfiles) / sizeof(DefaultProfiles[0]); i++) {

        const gchar *pipe = DefaultProfiles[i].pipe;

        // Has user installed support for this audio profile/audio format? Test pipe. 
        // Support for audio/video formats come from gstreamer1-plugins-xxxx packages. 
        gboolean pipe_OK = media_profile_test_pipeline(pipe);
        if (!pipe_OK) {
            // Test next format
            continue;
        }

        // Add this audio format to the combo

        const gchar *id = DefaultProfiles[i].id;
        const gchar *name = DefaultProfiles[i].name;
        const gchar *extension = DefaultProfiles[i].ext;

        // Typical audio format list is
        //  CD Quality, AAC (.m4a)
        //  CD Quality, Lossless (.flac)
        //  CD Quality, Lossy (.ogg)
        //  CD Quality, MP2 (.mp2)
        //  CD Quality, MP3 (.mp3)
        //  Voice, Lossless (.wav)
        //  Voice, Lossy (.spx)

        gchar *e = g_utf8_strup(extension, -1);
        gchar *n = g_utf8_strdown(name, -1);
        gchar *txt = g_strdup_printf(".%s  (%s)", e, n);

        // Add OGG type "CD Quality, Lossy" as first (becomes default)
        if (!g_strcmp0(extension, "ogg") || !g_strcmp0(extension, "oga")) {
            gtk_list_store_prepend(GTK_LIST_STORE(store), &iter);
        } else {
            // Add last
            gtk_list_store_append(GTK_LIST_STORE(store), &iter);
        }

        gtk_list_store_set(GTK_LIST_STORE(store), &iter, 0, id, 1, txt, -1);

        g_free(txt);
        g_free(e);
        g_free(n);
    }
}

static gboolean media_profiles_check_profile_EXT(gchar *id) {
    // Check if id is a valid profile ident
    const ProfileRec *rec = media_profile_get_EXT(id);
    if (rec) {
        return TRUE;
    }
    return FALSE;
}

static const ProfileRec *media_profile_get_EXT(gchar *id) {
    // Return ProfileRec for id
    guint i = 0;
    for (i = 0; i < sizeof(DefaultProfiles) / sizeof(DefaultProfiles[0]); i++) {
        if (!g_strcmp0(DefaultProfiles[i].id, id)) {
            return &DefaultProfiles[i];
        }
    }
    return NULL;
}

static gchar *media_profiles_get_pipeline_EXT(gchar *id) {
    // Return pipeline for id
    const ProfileRec *rec = media_profile_get_EXT(id);
    if (rec) {
        return g_strdup(rec->pipe);
    }
    return NULL;
}

static gchar *media_profiles_get_extension_EXT(gchar *id) {
    // Return file extension for id
    const ProfileRec *rec = media_profile_get_EXT(id);
    if (rec) {
        return g_strdup(rec->ext);
    }
    return NULL;
}


